/*
  GetElevation.cpp

  Implementation of the GetElevation function
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "TGAFile.h"

/*
long GetElevationUnscaled(const int x,
                          const int y)

Retrieve the pixel value (three byte signed integer) as a long integer.
m_zScale is not taken into account.
To retrieve the long integer of a pixel directly from the TGA file, see GetElevationFromFile().

Parameters:
x                      const int       x coordinate of pixel
                                       requested
y                      const int       y coordinate of pixel
                                       requested
Return value: the elevation (unscaled)
*/
long TGAFile::GetElevationUnscaled(const int x, const int y)
{
  if(!m_bInMemory)
  {
    return GetElevationFromFileUnscaled(x, y);
  }
  else
  {
    long lElevation = 0L;

    if((x >= 0) && (x < m_pHeader->image_width) && (y >= 0) && (y < m_pHeader->image_height))
    {
      int i;
      unsigned char* pPos1;
      unsigned char* pPos2;

      pPos1 = m_pData + y * m_dwWidthDWORDAligned + x * m_nBytesPerPixel;
      pPos2 = (unsigned char *)&lElevation;
      /* Put into pColor. */
      for(i = 0; i < m_nBytesPerPixel; i++)
      {
        *pPos2++ = *pPos1++; /* BLUE, GREEN, RED [, ALPHA]  */
      }
      if(lElevation >= 8388608)
      {
        lElevation -= 16777216;
      }
    }
  #ifdef _DEBUG
    else
    {
      fprintf(stderr,
              "Error: Overflow in TGAFile::GetElevation(%d, %d) - file size is only (%d, %d).\n",
              x, y, m_pHeader->image_width, m_pHeader->image_height);
    }
  #endif
    return lElevation;
  }
}

/*
double GetElevationScaled(const int x,
                          const int y)

Retrieve the pixel value (three byte signed integer) as a long integer.
m_zScale is used as the scaling factor.
To retrieve the long integer of a pixel directly from the TGA file, see GetElevationFromFile().

Parameters:
x                      const int       x coordinate of pixel
                                       requested
y                      const int       y coordinate of pixel
                                       requested
Return value: the elevation (unscaled)
*/
double TGAFile::GetElevationScaled(const int x, const int y)
{
  if(!m_bInMemory)
  {
    return GetElevationFromFileScaled(x, y);
  }
  else
  {
    long lElevation = 0L;

    if((x >= 0) && (x < m_pHeader->image_width) && (y >= 0) && (y < m_pHeader->image_height))
    {
      int i;
      unsigned char* pPos1;
      unsigned char* pPos2;

      pPos1 = m_pData + y * m_dwWidthDWORDAligned + x * m_nBytesPerPixel;
      pPos2 = (unsigned char *)&lElevation;
      /* Put into pColor. */
      for(i = 0; i < m_nBytesPerPixel; i++)
      {
        *pPos2++ = *pPos1++; /* BLUE, GREEN, RED [, ALPHA]  */
      }
      if(lElevation >= 8388608)
      {
        lElevation -= 16777216;
      }
    }
  #ifdef _DEBUG
    else
    {
      fprintf(stderr,
              "Error: Overflow in TGAFile::GetElevation(%d, %d) - file size is only (%d, %d).\n",
              x, y, m_pHeader->image_width, m_pHeader->image_height);
    }
  #endif
    return lElevation * m_zScale;
  }
}

/*
long GetElevationFromFileUnscaled(const int x,
                                  const int y)

Retrieve the pixel value (three byte signed integer) as a long integer directly from the TGA file.
m_zScale is not taken into account.

Parameters:
x                      const int       x coordinate of pixel
                                       requested
y                      const int       y coordinate of pixel
                                       requested
Return value: the elevation (unscaled)
*/
long TGAFile::GetElevationFromFileUnscaled(const int x, const int y)
{
  long lElevation = 0L;

  if((x >= 0) && (x < m_pHeader->image_width) && (y >= 0) && (y < m_pHeader->image_height))
  {
    fseek(m_pInputFile,
          (long)sizeof(TGAFILEHEADER) +
            (long)m_pHeader->ident_len +
            ((long)y * (long)m_pHeader->image_width +
            (long)x) * 3,
          SEEK_SET);
    fread(&lElevation, 3, 1, m_pInputFile);
    if(lElevation >= 8388608)
    {
      lElevation -= 16777216;
    }
  }
#ifdef _DEBUG
  else
  {
    fprintf(stderr,
            "Error: Overflow in TGAFile::GetElevationFromFile(%d, %d) - file size is only (%d, %d).\n",
            x, y, m_pHeader->image_width, m_pHeader->image_height);
  }
#endif
  return lElevation;
}

/*
double GetElevationFromFileScaled(const int x,
                                  const int y)

Retrieve the pixel value (three byte signed integer) as a long integer directly from the TGA file.
m_zScale is used as the scaling factor.

Parameters:
x                      const int       x coordinate of pixel
                                       requested
y                      const int       y coordinate of pixel
                                       requested
Return value: the elevation (scaled)
*/
double TGAFile::GetElevationFromFileScaled(const int x, const int y)
{
  long lElevation = 0L;

  if((x >= 0) && (x < m_pHeader->image_width) && (y >= 0) && (y < m_pHeader->image_height))
  {
    fseek(m_pInputFile,
          (long)sizeof(TGAFILEHEADER) +
            (long)m_pHeader->ident_len +
            ((long)y * (long)m_pHeader->image_width +
            (long)x) * 3,
          SEEK_SET);
    fread(&lElevation, 3, 1, m_pInputFile);
    if(lElevation >= 8388608)
    {
      lElevation -= 16777216;
    }
  }
#ifdef _DEBUG
  else
  {
    fprintf(stderr,
            "Error: Overflow in TGAFile::GetElevationFromFile(%d, %d) - file size is only (%d, %d).\n",
            x, y, m_pHeader->image_width, m_pHeader->image_height);
  }
#endif
  return lElevation * m_zScale;
}

/*
long GetElevationUnscaled(double x,
                          double y)

Retrieve the pixel value (three byte signed integer) as double
m_zScale is not taken into account.

Parameters:
x                      double          x coordinate of pixel
                                       requested
y                      double          y coordinate of pixel
                                       requested
Return value: the elevation (unscaled)

Global variables affected: none
*/
long TGAFile::GetElevationUnscaled(double x, double y)
{
  if(x < 0.0)
  {
    x = 0.0;
  }
  if(x > m_pHeader->image_width - 2.0)
  {
    x = m_pHeader->image_width - 2.0;
  }
  if(y < 0.0)
  {
    y = 0.0;
  }
  if(y > m_pHeader->image_height - 2.0)
  {
    y = m_pHeader->image_height - 2.0;
  }

  int xint = (int)x;
  int yint = (int)y;
  double xrest   = x - floor(x);
  double yrest   = y - floor(y);
  double x1_rest = 1.0 - xrest;
  double y1_rest = 1.0 - yrest;

  double dElevation[4];

  dElevation[0] = GetElevationUnscaled(xint,     yint    );
  dElevation[1] = GetElevationUnscaled(xint + 1, yint    );
  dElevation[2] = GetElevationUnscaled(xint,     yint + 1);
  dElevation[3] = GetElevationUnscaled(xint + 1, yint + 1);

  double dE = yrest * (x1_rest * dElevation[2] + xrest * dElevation[3]) + y1_rest * (x1_rest * dElevation[0] + xrest * dElevation[1]);

  return long(dE + 0.5);
}

/*
double GetElevationScaled(double x,
                          double y)

Retrieve the pixel value (three byte signed integer) as double
m_zScale is used as the scaling factor.

Parameters:
x                      double          x coordinate of pixel
                                       requested
y                      double          y coordinate of pixel
                                       requested
Return value: the elevation (scaled)

Global variables affected: none
*/
double TGAFile::GetElevationScaled(double x, double y)
{
  if(x < 0.0)
  {
    x = 0.0;
  }
  if(x > m_pHeader->image_width - 2.0)
  {
    x = m_pHeader->image_width - 2.0;
  }
  if(y < 0.0)
  {
    y = 0.0;
  }
  if(y > m_pHeader->image_height - 2.0)
  {
    y = m_pHeader->image_height - 2.0;
  }

  int xint = (int)x;
  int yint = (int)y;
  double xrest   = x - floor(x);
  double yrest   = y - floor(y);
  double x1_rest = 1.0 - xrest;
  double y1_rest = 1.0 - yrest;

  double dElevation[4];

  dElevation[0] = GetElevationUnscaled(xint,     yint    ); // Don't scale here! Only in the end.
  dElevation[1] = GetElevationUnscaled(xint + 1, yint    );
  dElevation[2] = GetElevationUnscaled(xint,     yint + 1);
  dElevation[3] = GetElevationUnscaled(xint + 1, yint + 1);

  double dE = yrest * (x1_rest * dElevation[2] + xrest * dElevation[3]) + y1_rest * (x1_rest * dElevation[0] + xrest * dElevation[1]);

  return dE * m_zScale; // Here, do the scaling.
}
