// BilToTga.cpp: implementation file

#include "stdafx.h"

#include <afxtempl.h>
#include <float.h>

#include <list>
#include "Waypoint.h"

#include "geoDataClient.h"
#include "geoDataClientDialog.h"
#include "HTTPRequest.h"

// Convert the BIL file to a TGA file using the HDR file.
// Returns 0 if OK.
int CgeoDataClientDialog::BilToTga(const CString &bilFilename,
                                   const CString &hdrFilename,
                                   const CString &stxFilename,
                                   const CString &tgaFilename)
{
  nBILRows = 0;
  nBILCols = 0;
  int nBitsPerPixel = 0;
  int nBytesPerPixel = 0;
  double dZScale = 1.0;
  long lNoData = -9999;

  CString message;
  message.Format(_T("Converting \"%s\" to \"%s\"."), bilFilename, tgaFilename);
  addMessage(message);

  enum
  {
    Motorola,
    Intel
  } eByteorder = Intel;

  CStdioFile headerFile;
  if(headerFile.Open(hdrFilename, CFile::modeRead | CFile::typeText))
  {
    CString line;
    while(headerFile.ReadString(line))
    {
      int curPos = 0;
      CString key = line.Tokenize(" ", curPos);
      CString value;
      if(!key.IsEmpty())
      {
        value = line.Tokenize(" ", curPos);
      }
      if(!key.CompareNoCase("NROWS"))
      {
        nBILRows = atoi(value);
      }
      if(!key.CompareNoCase("NCOLS"))
      {
        nBILCols = atoi(value);
      }
      if(!key.CompareNoCase("NBITS"))
      {
        nBitsPerPixel = atoi(value);
        nBytesPerPixel = nBitsPerPixel / 8;
      }
      if(!key.CompareNoCase("BYTEORDER"))
      {
        if((value[0] == 'm') || (value[0] == 'M'))
        {
          eByteorder = Motorola;
        }
      }
    }
    headerFile.Close();
  }
  else
  {
    message.Format(_T("Error in BilToTga(): Could not open \"%s\"."), hdrFilename);
    addMessage(message);
    return 1; // Error.
  }

  int one = 1;
  int minimum = -9999;
  int maximum =  9999;
  double mean = 0.0;
  double stddev = 0.0;

  CStdioFile statisticsFile;
  if(statisticsFile.Open(stxFilename, CFile::modeRead | CFile::typeText))
  {
    CString line;
    statisticsFile.ReadString(line);
    // The STX file (statistics) consists of a 1, then the minimum,
    // maximum, mean, and standard deviation of the Z values.
    // The -74, -73, 41, 40 quad has NBITS 8! This is that quad's STX file:
    // 1 -56 197 8.9 18.5
    char strMean[1024];
    char strStddev[1024];
    if(sscanf(line, "%d %d %d %s %s", &one, &minimum, &maximum, strMean, strStddev) != 5)
//    if(sscanf(line, "%d %d %d %lf %lf", &one, &minimum, &maximum, &mean, &stddev) != 5)
    {
      message.Format(_T("Error in BilToTga(): Could not read data from \"%s\"."), stxFilename);
      addMessage(message);
      return 1; // Error.
    }
	switch(minimum)
	{
		case -9999:
		case -32768:
		{
			lNoData = minimum;
		}
	}
    mean = atof(strMean);
    stddev = atof(strStddev);
    statisticsFile.Close();
  }
  else
  {
    message.Format(_T("Error in BilToTga(): Could not open \"%s\"."), stxFilename);
    addMessage(message);
    return 1; // Error.
  }
  CFile bilFile;
  if(bilFile.Open(bilFilename, CFile::modeRead | CFile::typeBinary))
  {
	char first4Bytes[4];
	bilFile.Read(first4Bytes, 4);
	const char *magicWord = "\x49\x49\x2A\x00";
	LONGLONG seekOffset = 0;
	if(!memcmp(first4Bytes, magicWord, 4))
	{
		// 8/27/2009: the new format with payload data embedded between other information.
		char *firstMegabyte = new char[1000000 - 4];
		bilFile.Read(firstMegabyte, 1000000 - 4);
		const char *WGS84 = "\x23\x57\x47\x53\x20\x38\x34\x7C\x00";
		int pos;
		for(pos = 0; pos < 1000000 - 4 - 9; pos++)
		{
			if(!memcmp(firstMegabyte + pos, WGS84, 9))
			{
				break;
			}
		}
		delete firstMegabyte;
		firstMegabyte = 0;
		if(pos < 1000000 - 4 - 9)
		{
			seekOffset = 4 + pos + 9;
			// Found.
		}
		else
		{
			if(MessageBox(_T("The BIL file begins with \"II*\", which requires the \"WGS 84\" keyword\n"
							 "to locate the beginning of the elevation data block in the BIL file.\n"
							 "That keyword could not be located. An import can only be attempted\n"
							 "beginning at the first byte, most probably leading to incorrect results.\n"
							 "\n"
							 "Proceed anyway?"),
						  _T("Warning"),
						  MB_YESNO | MB_ICONINFORMATION) == IDNO)
			{
				return 1;
			}
		}
	}
    TGAFile *tgaFile = new TGAFile();
    tgaFile->New(nBILCols - 1, nBILRows - 1);
    // Write the elevation data
    long lTotalRowBytes = nBILCols * nBytesPerPixel;
    // szValue: Bytes in the input file: high and low order byte.
    unsigned char *szValue = new unsigned char[lTotalRowBytes];
    // pValue: Pointer to input values.
    unsigned char *pValue = 0;
    // lRow: Vertical starting position.
    long lRow = nBILRows - 1;
    // lCol: Horizontal starting position.
    long lCol = 0;

    // F32: a pixel in a file with 32 bit per pixel (float).
    union
    {
      float f;
      unsigned char c[4];
    } F32;
    // I16: a pixel in a file with 16 bit per pixel (integer).
    union
    {
      short int i;
      unsigned char c[2];
    } I16;
    // I32: a pixel in a file with 32 bit per pixel (integer).
    union
    {
      long l;
      unsigned char c[4];
    } I32;
    unsigned char cValue;
    double dValue;  
    double dMinValue =  FLT_MAX; // Only for debugging and diagnostics.
    double dMaxValue = -FLT_MAX; // Only for debugging and diagnostics.

    RGBQUAD Color;
    unsigned char *pColor = (unsigned char*)&Color;
    pColor[3] = 0; // The fourth byte is always zero.

    for(int y = 0; y < nBILRows - 1; y++)
    {
      bilFile.Seek(seekOffset + lRow * lTotalRowBytes + nBytesPerPixel * lCol, CFile::begin);
      bilFile.Read(szValue, nBytesPerPixel * nBILCols);
      pValue = szValue;
      switch(nBytesPerPixel)
      {
        case 1:
          {
            for(int x = 0; x < nBILCols - 1; x++)
            {
              cValue = *pValue++;
              if(cValue > maximum) // cValue is of type unsigned char
              {
                dValue = ((double)(cValue) - 256.0) * dZScale;
              }
              else
              {
                dValue = (double)(cValue) * dZScale;
              }
              if(dValue >  8388607.0)
              {
                dValue =  8388607.0;
              }
              if(dValue < -8388608.0)
              {
                dValue = -8388608.0;
              }
              I32.l = (long)(dValue + 0.5);
              pColor = (unsigned char*)&Color;
              *pColor++ = I32.c[0];  // blue
              *pColor++ = I32.c[1];  // green
              *pColor   = I32.c[2];  // red
              if(dValue > dMaxValue)
              {
                dMaxValue = dValue;
              }
              if(dValue < dMinValue)
              {
                dMinValue = dValue;
              }
              tgaFile->SetColor(x, y, &Color);
            }
            break;
          }
        case 2:
          {
            if(eByteorder == Motorola)
            {
              for(int x = 0; x < nBILCols - 1; x++)
              {
                I16.c[1] = *pValue++;
                I16.c[0] = *pValue++;
                if(I16.i == lNoData)
                {
                  I16.i = 0;
                }
                dValue = (double)(I16.i) * dZScale;
                if(dValue >  8388607.0)
                {
                  dValue =  8388607.0;
                }
                if(dValue < -8388608.0)
                {
                  dValue = -8388608.0;
                }
                I32.l = (long)(dValue + 0.5);
                pColor = (unsigned char*)&Color;
                *pColor++ = I32.c[0];  // blue
                *pColor++ = I32.c[1];  // green
                *pColor   = I32.c[2];  // red
                if(dValue > dMaxValue)
                {
                  dMaxValue = dValue;
                }
                if(dValue < dMinValue)
                {
                  dMinValue = dValue;
                }
                tgaFile->SetColor(x, y, &Color);
              }
            }
            else
            {
              for(int x = 0; x < nBILCols - 1; x++)
              {
                I16.c[0] = *pValue++;
                I16.c[1] = *pValue++;
                if(I16.i == lNoData)
                {
                  I16.i = 0;
                }
                dValue = (double)(I16.i) * dZScale;
                if(dValue >  8388607.0)
                {
                  dValue =  8388607.0;
                }
                if(dValue < -8388608.0)
                {
                  dValue = -8388608.0;
                }
                I32.l = (long)(dValue + 0.5);
                pColor = (unsigned char*)&Color;
                *pColor++ = I32.c[0];  // blue
                *pColor++ = I32.c[1];  // green
                *pColor   = I32.c[2];  // red
                if(dValue > dMaxValue)
                {
                  dMaxValue = dValue;
                }
                if(dValue < dMinValue)
                {
                  dMinValue = dValue;
                }
                tgaFile->SetColor(x, y, &Color);
              }
            }
            break;
          }
        case 4:
          {
            if(eByteorder == Motorola)
            {
              for(int x = 0; x < nBILCols - 1; x++)
              {
                F32.c[3] = *pValue++;
                F32.c[2] = *pValue++;
                F32.c[1] = *pValue++;
                F32.c[0] = *pValue++;
                if(F32.f == (float)lNoData)
                {
                  F32.f = (float)0.0;
                }
                dValue = (double)(F32.f) * dZScale;
                if(dValue >  8388607.0)
                {
                  dValue =  8388607.0;
                }
                if(dValue < -8388608.0)
                {
                  dValue = -8388608.0;
                }
                I32.l = (long)(dValue + 0.5);
                pColor = (unsigned char*)&Color;
                *pColor++ = I32.c[0];  // blue
                *pColor++ = I32.c[1];  // green
                *pColor   = I32.c[2];  // red
                if(dValue > dMaxValue)
                {
                  dMaxValue = dValue;
                }
                if(dValue < dMinValue)
                {
                  dMinValue = dValue;
                }
                tgaFile->SetColor(x, y, &Color);
              }
            }
            else
            {
              for(int x = 0; x < nBILCols - 1; x++)
              {
                F32.c[0] = *pValue++;
                F32.c[1] = *pValue++;
                F32.c[2] = *pValue++;
                F32.c[3] = *pValue++;
                if(F32.f == (float)lNoData)
                {
                  F32.f = (float)0.0;
                }
                dValue = (double)(F32.f) * dZScale;
                if(dValue >  8388607.0)
                {
                  dValue =  8388607.0;
                }
                if(dValue < -8388608.0)
                {
                  dValue = -8388608.0;
                }
                I32.l = (long)(dValue + 0.5);
                pColor = (unsigned char*)&Color;
                *pColor++ = I32.c[0];  // blue
                *pColor++ = I32.c[1];  // green
                *pColor   = I32.c[2];  // red
                if(dValue > dMaxValue)
                {
                  dMaxValue = dValue;
                }
                if(dValue < dMinValue)
                {
                  dMinValue = dValue;
                }
                tgaFile->SetColor(x, y, &Color);
              }
            }
            break;
          }
      }
      lRow--;
    }
    bilFile.Close();
    delete szValue;
    szValue = 0;
    tgaFile->Write(tgaFilename);
    message.Format(_T("BIL import successful: \"%s\" written."), tgaFilename);
    addMessage(message);
    delete tgaFile;
    tgaFile = 0;
    // Note: nBILCols is one more than the BIL file width, the last column already belongs to the next tile.
    nBILCols--; // The rightmost column is not counted.
    // Note: nBILRows is one more than the BIL file height, the last row already belongs to the next tile.
    nBILRows--; // The bottom row is not counted.
    return 0; // No error.
  }
  else
  {
    message.Format(_T("Error in BilToTga(): Could not open \"%s\"."), bilFilename);
    addMessage(message);
    return 1; // Error.
  }
}
