// GifToTga.cpp: implementation file

#include "stdafx.h"

#include <afxtempl.h>
#include <float.h>

#include <list>
#include "Waypoint.h"

#include "geoDataClient.h"
#include "geoDataClientDialog.h"
#include "HTTPRequest.h"

// Convert a GIF file to a TGA file, scale 1:1
// Returns 0 if OK.
int CgeoDataClientDialog::GifToTga(const CString &gifFilename,
                                   const CString &tgaFilename)
{
  CString message;
  message.Format(_T("Converting \"%s\" to \"%s\"."), gifFilename, tgaFilename);
  addMessage(message);

  gl_ZeroDataBlock = 0;
  gl_nShowComment  = 1; // flag to show comments
  gl_nVerbose      = 0; // flag for verbose output
  TGAFile tgaFile;
  if(ReadGIFFile(gifFilename, 1, &tgaFile))
  {
    message.Format("Error in GifToTga(): Cannot read GIF file \"%s\".", gifFilename);
    addMessage(message);
    return 2;
  }
  if(tgaFile.Write(tgaFilename))
  {
    message.Format(_T("Error in GifToTga(): Cannot write TGA file \"%s\"."), tgaFilename);
    addMessage(message);
    return 1;
  }
  message.Format(_T("GIF conversion successful: \"%s\" written."), tgaFilename);
  addMessage(message);
  return 0;
}

/*
int ReadGIFFile(const char *szFilename,
                int nImageNumber,
                TGAFile *pTGAFile)

Parameters:
szFilename    const char *  Name of input file
nImageNumber  int           Number of interlaced image to retrieve
pTGAFile      TGAFile *     Pointer to TGA file

Return value:
1 if an error occurred
0 if no error occurred

Global variables affected:
gl_pHeader
gl_lWidthDWORDAligned
gl_pData
*/
int CgeoDataClientDialog::ReadGIFFile(const char *szFilename,
                                      int nImageNumber,
                                      TGAFile *pTGAFile)
{
  unsigned char  szBuffer[256];
  GIFSCREEN      GifScreen;
  unsigned char  c;
  int            nImageCount;
  int            nUseGlobalColormap;
  int            nBitPixel;
  unsigned char  LocalColorMap[3][MAXCOLORMAPSIZE];
  tFileFormat    FileFormat;
  CString        message;

  FILE *pGIFFile;
  pGIFFile = fopen(szFilename, "rb");
  if(!pGIFFile)
  {
    return 1;
  }
  /* The first 6 bytes have to be either "GIF87a" or "GIF89a" */
  if(fread(szBuffer, 1, 6, pGIFFile) != 6)
  {
    fclose(pGIFFile);
    message.Format("Error in ReadGIFFile(): Failed to read magic number.");
    addMessage(message);
    return 2;
  }
  if(strncmp((char *)szBuffer,
             "GIF",
             3) != 0)
  {
    fclose(pGIFFile);
    addMessage(_T("Error in ReadGIFFile(): Not a GIF file."));
    return 3;
  }
  if((strncmp((char *)(szBuffer + 3),
              "87a",
              3) != 0) &&
     (strncmp((char *)(szBuffer + 3),
             "89a",
             3) != 0))
  {
    fclose(pGIFFile);
    message.Format("Error in ReadGIFFile(): Not a 87a or 89a GIF file.");
    addMessage(message);
    return 4;
  }
  if(fread(szBuffer, 1, 7, pGIFFile) != 7)
  {
    fclose(pGIFFile);
    addMessage(_T("Error in ReadGIFFile(): Failed to read screen descriptor."));
    return 5;
  }
  GifScreen.Width           = szBuffer[0] + (szBuffer[1] << 8);
  GifScreen.Height          = szBuffer[2] + (szBuffer[3] << 8);
  GifScreen.BitPixel        = 2 << (szBuffer[4] & 0x07);
  GifScreen.ColorResolution = (((szBuffer[4] & 0x70) >> 3) + 1);
  GifScreen.Background      = szBuffer[5];
  GifScreen.AspectRatio     = szBuffer[6];

  if(BitSet(szBuffer[4], LOCALCOLORMAP))
  {
    /* Global Colormap */
    if(ReadColorMap(pGIFFile,
                    GifScreen.BitPixel,
                    GifScreen.ColorMap,
                    &GifScreen.FileFormat))
    {
      fclose(pGIFFile);
      addMessage(_T("Error in ReadGIFFile(): Error reading global colormap."));
      return 6;
    }
  }

  if(GifScreen.AspectRatio != 0 &&
     GifScreen.AspectRatio != 49)
  {
    double dAspectRatio;
    dAspectRatio = ((double)GifScreen.AspectRatio + 15.0) / 64.0;
    addMessage(_T("Error in ReadGIFFile(): Warning - non-square pixels"));
    message.Format(_T("Error in ReadGIFFile(): To fix do a 'pnmscale -%cscale %g'."),
                   dAspectRatio < 1.0 ? 'x' : 'y',
                   dAspectRatio < 1.0 ? 1.0 / dAspectRatio : dAspectRatio);
    addMessage(message);
  }

  pTGAFile->New(GifScreen.Width, GifScreen.Height);

  nImageCount = 0;
  for(;;)
  {
    if(fread(&c, 1, 1, pGIFFile) != 1)
    {
      fclose(pGIFFile);
      message.Format("Error in ReadGIFFile(): EOF/read error on image data.");
      addMessage(message);
      return 7;
    }
    if(c == ';')
    {
      /* GIF terminator */
      if(nImageCount < nImageNumber)
      {
        fclose(pGIFFile);
        message.Format(_T("Error in ReadGIFFile(): Only %d image%s found in file."),
                       nImageCount,
                       nImageCount > 1 ? "s" : "");
        addMessage(message);
        return 8;
      }
      fclose(pGIFFile);
      return 0;
    }

    if(c == '!')
    {
      /* Extension */
      if(fread(&c, 1, 1, pGIFFile) != 1)
      {
        fclose(pGIFFile);
        addMessage(_T("Error in ReadGIFFile(): OF/read error on extention function code."));
        return 9;
      }
      DoExtension(pGIFFile, c);
      continue;
    }

    if(c != ',')
    {         /* Not a valid start character */
      fclose(pGIFFile);
      message.Format(_T("ReadGIFFile(): Bogus character 0x%02x, ignoring."), (int)c);
      addMessage(message);
      continue;
    }

    nImageCount++;

    if(fread(szBuffer, 1, 9, pGIFFile) != 9)
    {
      fclose(pGIFFile);
      addMessage(_T("Error in ReadGIFFile(): Couldn't read left/top/width/height."));
      return 10;
    }
    nUseGlobalColormap = !BitSet(szBuffer[8], LOCALCOLORMAP);

    nBitPixel = 1 << ((szBuffer[8] & 0x07) + 1);

    if(!nUseGlobalColormap)
    {
      if(ReadColorMap(pGIFFile,
                      nBitPixel,
                      LocalColorMap,
                      &FileFormat))
      {
        fclose(pGIFFile);
        addMessage(_T("Error in ReadGIFFile(): Error reading local colormap."));
        return 11;
      }
      ReadImage(pGIFFile,
                szBuffer[4] + (szBuffer[5] << 8),
                szBuffer[6] + (szBuffer[7] << 8), 
                LocalColorMap,
                FileFormat,
                BitSet(szBuffer[8], INTERLACE),
                nImageCount != nImageNumber,
                pTGAFile);
    }
    else
    {
      ReadImage(pGIFFile,
                szBuffer[4] + (szBuffer[5] << 8),
                szBuffer[6] + (szBuffer[7] << 8),
                GifScreen.ColorMap,
                GifScreen.FileFormat,
                BitSet(szBuffer[8], INTERLACE),
                nImageCount != nImageNumber,
                pTGAFile);
    }
  }
  fclose(pGIFFile);
  return 0;
}

/*
int ReadColorMap(FILE *pInputfile,
                 int nColorCount,
                 unsigned char buffer[3][MAXCOLORMAPSIZE],
                 tFileFormat *pFileFormat)

Parameters:
pInputfile             FILE *          Pointer to input file
nColorCount            int             Number of colors in the map
buffer[3][MAXCOLORMAPSIZE]
                       unsigned char   Color map array
pFileFormat            tFileFormat *   Format: B/W, grayscale, color

Return value:
1 if an error occurred
0 if no error occurred

Global variables affected:
gl_pHeader
gl_lWidthDWORDAligned
gl_pData
*/
int CgeoDataClientDialog::ReadColorMap(FILE *pInputfile,
                                       int nColorCount,
                                       unsigned char buffer[3][MAXCOLORMAPSIZE],
                                       tFileFormat *FileFormat)
{
  int     nColor;
  RGBQUAD Color;
  CString message;

  if(nColorCount == 2)
  {
    *FileFormat = FileFormatMonochrome;
  }
  else
  {
    *FileFormat = FileFormatGrayscale;
  }
  for(nColor = 0; nColor < nColorCount; nColor++)
  {
    if(fread(&Color, 1, 3, pInputfile) != 3)
    {
      addMessage(_T("Error in ReadColorMap(): Bad colormap."));
      return 1;
    }

    buffer[CM_RED][nColor]   = Color.rgbBlue;
    buffer[CM_GREEN][nColor] = Color.rgbGreen;
    buffer[CM_BLUE][nColor]  = Color.rgbRed;

    if(Color.rgbRed != Color.rgbGreen ||
       Color.rgbGreen != Color.rgbBlue)
    {
      /* The very first occurrence of different R, G, B values
         makes the format colored. */
      *FileFormat = FileFormatColor;
    }
  }
  return 0;
}

int CgeoDataClientDialog::DoExtension(FILE *pInputfile,
                                      int nLabel)
{
  static char     szBuffer[256];
  char            *szString;
  CString         message;

  switch(nLabel)
  {
    case 0x01:              /* Plain Text Extension */
               szString = "Plain Text Extension";
               break;
    case 0xff:              /* Application Extension */
               szString = "Application Extension";
               break;
    case 0xfe:              /* Comment Extension */
               szString = "Comment Extension";
               while(GetDataBlock(pInputfile,
                                  (unsigned char*)szBuffer) != 0)
               {
                 if(gl_nShowComment)
                 {
                   message.Format(_T("DoExtension(): GIF comment: %s."),
                                  szBuffer);
                   addMessage(message);
                 }
               }
               return 0;
    case 0xf9:              /* Graphic Control Extension */
               szString = "Graphic Control Extension";
               GetDataBlock(pInputfile, (unsigned char*)szBuffer);
               Gif89.disposal    = (szBuffer[0] >> 2) & 0x7;
               Gif89.inputFlag   = (szBuffer[0] >> 1) & 0x1;
               Gif89.delayTime   = szBuffer[1] + (szBuffer[2] << 8);
               if((szBuffer[0] & 0x1) != 0)
               {
                 Gif89.transparent = szBuffer[3];
               }

               while(GetDataBlock(pInputfile, (unsigned char*)szBuffer) != 0)
               {
                 /* do nothing */
               }
               return 0;
       default:
               szString = szBuffer;
               sprintf(szBuffer,
                       "UNKNOWN (0x%02x)",
                       nLabel);
               break;
       }

       message.Format(_T("DoExtension(): Got a \"%s\" extension."), szString);
       addMessage(message);

       while (GetDataBlock(pInputfile,
                          (unsigned char*)szBuffer) != 0)
       {
         /* do nothing */
       }

       return 0;
}

void CgeoDataClientDialog::ReadImage(FILE *pInputfile,
                                     int nWidth,
                                     int nHeight,
                                     unsigned char cmap[3][MAXCOLORMAPSIZE],
                                     tFileFormat FileFormat,
                                     int interlace,
                                     int ignore,
                                     TGAFile *pTGAFile)
{
  unsigned char   c;      
  int             v;
  int             xpos = 0, ypos = 0, pass = 0;
  RGBQUAD         Color;
  CString         message;
  Color.rgbReserved = 0;

  /*
  **  Initialize the Compression routines
  */
  if(fread(&c, 1, 1, pInputfile) != 1)
  {
    addMessage(_T("Error in ReadImage(): EOF/read error on image data."));
    return;
  }
  if(LWZReadByte(pInputfile, 1, c) < 0)
  {
    addMessage(_T("Error in ReadImage(): Error reading image."));
    return;
  }

  /*
  **  If this is an "uninteresting picture" ignore it.
  */
  if(ignore)
  {
    if(gl_nVerbose)
    {
      addMessage(_T("ReadImage(): Skipping image..."));
    }

    while(LWZReadByte(pInputfile, 0, c) >= 0)
    {
      /* do nothing */
    }
    return;
  }

  if(gl_nVerbose)
  {
    message.Format(_T("ReadImage(): Reading %d by %d%s GIF image."),
                   nWidth,
                   nHeight,
                   interlace ? " interlaced" : "");
    addMessage(message);
  }
  while((v = LWZReadByte(pInputfile, 0, c)) >= 0)
  {
    Color.rgbRed   = cmap[CM_RED][v];
    Color.rgbGreen = cmap[CM_GREEN][v];
    Color.rgbBlue  = cmap[CM_BLUE][v];
    pTGAFile->SetColor(xpos,
              pTGAFile->GetImageHeight() - ypos - 1,
              &Color);

    ++xpos;
    if(xpos == nWidth)
    {
      xpos = 0;
      if(interlace)
      {
        switch(pass)
        {
          case 0:
          case 1:
                  ypos += 8; break;
          case 2:
                  ypos += 4; break;
          case 3:
                  ypos += 2; break;
        }
        if(ypos >= nHeight)
        {
          ++pass;
          switch(pass)
          {
            case 1:
                    ypos = 4; break;
            case 2:
                    ypos = 2; break;
            case 3:
                    ypos = 1; break;
            default:
                    goto fini;
          }
        }
      }
      else
      {
        ++ypos;
      }
    }
    if(ypos >= nHeight)
    {
      break;
    }
  }

fini:
  if(LWZReadByte(pInputfile, 0, c) >= 0)
  {
    addMessage(_T("ReadImage(): Too much input data, ignoring extra..."));
  }

  if(gl_nVerbose)
  {
    message.Format(_T("ReadImage(): Writing a %s file."),
                   FileFormat == FileFormatMonochrome ? "monochrome" :
                   FileFormat == FileFormatGrayscale  ? "grayscale" :
                   FileFormat == FileFormatColor      ? "color" :
                   "unknown");
    addMessage(message);
  }

  /*
  pnm_writepnm(stdout, image, len, height, (pixval) 255, pbm_format, 0 );
  */
}

int CgeoDataClientDialog::GetDataBlock(FILE *pInputfile,
                          unsigned char *pBuffer)
{
  unsigned char cCount;
  CString       message;

  if(fread(&cCount, 1, 1, pInputfile) != 1)
  {
    addMessage(_T("Error in GetDataBlock(): Error in getting DataBlock size."));
    return -1;
  }

  gl_ZeroDataBlock = (cCount == 0);

  if((cCount != 0) &&
     (fread(pBuffer,
            1,
            cCount,
            pInputfile) != cCount))
  {
    addMessage(_T("Error in GetDataBlock(): Error in reading DataBlock."));
    return -1;
  }

  return cCount;
}

int CgeoDataClientDialog::LWZReadByte(FILE *pInputfile,
                                      int nFlag,
                                      int nInputCodeSize)
{
  static int      fresh = 0;
  int             code, incode;
  static int      code_size, set_code_size;
  static int      max_code, max_code_size;
  static int      firstcode, oldcode;
  static int      clear_code, end_code;
  static int      table[2][(1<< MAX_LWZ_BITS)];
  static int      stack[(1<<(MAX_LWZ_BITS))*2], *sp;
  register int    i;
  CString         message;

  if(nFlag)
  {
    set_code_size = nInputCodeSize;
    code_size     = set_code_size + 1;
    clear_code    = 1 << set_code_size ;
    end_code      = clear_code + 1;
    max_code_size = 2 * clear_code;
    max_code      = clear_code + 2;

    GetCode(pInputfile, 0, 1);
               
    fresh = 1;

    for (i = 0; i < clear_code; ++i)
    {
      table[0][i] = 0;
      table[1][i] = i;
    }
    for (; i < (1<<MAX_LWZ_BITS); ++i)
    {
      table[0][i] = table[1][0] = 0;
    }

    sp = stack;

    return 0;
  }
  else if(fresh)
  {
    fresh = 0;
    do
    {
      firstcode = oldcode = GetCode(pInputfile, code_size, 0);
    } while(firstcode == clear_code);
    return firstcode;
  }

  if(sp > stack)
  {
    return *--sp;
  }

  while((code = GetCode(pInputfile, code_size, 0)) >= 0)
  {
    if(code == clear_code)
    {
      for(i = 0; i < clear_code; ++i)
      {
        table[0][i] = 0;
        table[1][i] = i;
      }
      for(; i < (1<<MAX_LWZ_BITS); ++i)
      {
        table[0][i] = table[1][i] = 0;
      }
      code_size = set_code_size+1;
      max_code_size = 2 * clear_code;
      max_code = clear_code + 2;
      sp = stack;
      firstcode = oldcode = GetCode(pInputfile, code_size, 0);
      return firstcode;
    }
    else
      if(code == end_code)
      {
        int             count;
        unsigned char   buf[260];

        if(gl_ZeroDataBlock)
          return -2;

        while((count = GetDataBlock(pInputfile, buf)) > 0)
        {
          /* do nothing */
        }

        if(count != 0)
        {
          addMessage(_T("LWZReadByte(): Missing EOD in data stream (common occurence)."));
        }
        return -2;
      }

    incode = code;

    if(code >= max_code)
    {
      *sp++ = firstcode;
      code = oldcode;
    }

    while(code >= clear_code)
    {
      *sp++ = table[1][code];
      if(code == table[0][code])
      {
        addMessage(_T("Error in LWZReadByte(): Circular table entry (BIG ERROR)."));
      }
      code = table[0][code];
    }

    *sp++ = firstcode = table[1][code];

    if((code = max_code) <(1<<MAX_LWZ_BITS))
    {
      table[0][code] = oldcode;
      table[1][code] = firstcode;
      ++max_code;
      if((max_code >= max_code_size) &&
         (max_code_size < (1<<MAX_LWZ_BITS)))
      {
        max_code_size *= 2;
        ++code_size;
      }
    }

    oldcode = incode;

    if(sp > stack)
      return *--sp;
  }
  return code;
}

int CgeoDataClientDialog::GetCode(FILE *pInputfile,
                                  int nCodeSize,
                                  int nFlag)
{
  static unsigned char    buf[280];
  static int              curbit, lastbit, done, last_byte;
  int                     i, j, ret;
  unsigned char           count;
  CString                 message;

  if(nFlag)
  {
    curbit  = 0;
    lastbit = 0;
    done    = 0;
    return 0;
  }

  if((curbit + nCodeSize) >= lastbit)
  {
    if(done)
    {
      if(curbit >= lastbit)
      {
        addMessage(_T("GetCode(): Ran off the end of my bits."));
      }
      return -1;
    }
    buf[0] = buf[last_byte-2];
    buf[1] = buf[last_byte-1];

    if((count = GetDataBlock(pInputfile, &buf[2])) == 0)
      done = 1;

    last_byte = 2 + count;
    curbit = (curbit - lastbit) + 16;
    lastbit = (2+count)*8 ;
  }

  ret = 0;
  for(i = curbit, j = 0;
      j < nCodeSize;
      ++i, ++j)
  {
    ret |= ((buf[ i / 8 ] & (1 << (i % 8))) != 0) << j;
  }

  curbit += nCodeSize;

  return ret;
}
