// InterpolateElevation.cpp: implementation file

#include "stdafx.h"

#include <math.h>

#include <list>
#include "Waypoint.h"

#include "geoDataClient.h"
#include "geoDataClientDialog.h"

// Interpolate zero elevation values.
// Returns 0 if OK.
int CgeoDataClientDialog::interpolateElevation(const CString &inputFilename,
                                               int repeats)
{
	const double weight = 0.1; // The participation of the current loop in the whole image.
	const double weightComplement = 1.0 - weight;
  const CString outputFilename = inputFilename + "-int.tga";

  CString message;
  addMessage(_T("Substituting NODATA (z = 0) values via 2D relaxation."));
  TGAFile inputFile;
  if(inputFile.Open(inputFilename))
  {
    message.Format(_T("Could not open \"%s\"."), inputFilename);
    addMessage(message);
    return 1;
  }
  TGAFile outputFile;
  if(outputFile.New(inputFile.GetImageWidth(), inputFile.GetImageHeight()))
  {
    message.Format(_T("Could not allocate space for \"%s\"."), outputFilename);
    addMessage(message);
    return 2;
  }
	for(int y = 0; y < inputFile.GetImageHeight(); y++)
	{
		for(int x = 0; x < inputFile.GetImageWidth(); x++)
		{
      outputFile.SetElevationUnscaled(x, y, inputFile.GetElevationUnscaled(x, y));
    }
  }
	long oldResidual = 0;
	for(int repeat = 0; repeat < repeats; repeat++)
	{
    if(bCancel)
    {
      break;
    }
		long residual = 0;
		for(int y = 0; y < inputFile.GetImageHeight(); y++)
		{
			for(int x1 = 0; x1 < inputFile.GetImageWidth(); x1++)
			{
				long elevation1 = inputFile.GetElevationUnscaled(x1, y);
				if(elevation1 == 0)
				{
					// The elevation is zero. Find next non-zero elevation.
					for(int x2 = x1 + 1; x2 < inputFile.GetImageWidth(); x2++)
					{
						long elevation2 = inputFile.GetElevationUnscaled(x2, y);
						if(elevation2 != 0)
						{
							// The elevation is non-zero. Interpolate.
							// Get last non-zero elevation before the gap.
							long last;
							if(x1 > 0)
							{
								last = inputFile.GetElevationUnscaled(x1 - 1, y);
							}
							else
							{
 								last = elevation2;
 							}
							// elevation2 is the first non-zero elevation after the gap.
							// Interpolate between last and elevation2.
							double dX = x2 - x1 + 1; // The length of the gap.
							double elevation3 = last;
							double dElevation = (elevation2 - elevation3) / dX;
							for(int x3 = x1; x3 < x2; x3++)
							{
								elevation3 += dElevation;
                long elevationExisting = outputFile.GetElevationUnscaled(x3, y);
								long mixElevation = long(elevation3 * weight + elevationExisting * weightComplement + 0.5);
								residual += abs(mixElevation - elevationExisting);
								outputFile.SetElevationUnscaled(x3, y, mixElevation);
							}
							x1 = x2;
							break;
						}
					}
				}
			}
		}
		for(int x = 0; x < inputFile.GetImageWidth(); x++)
		{
			for(int y1 = 0; y1 < inputFile.GetImageHeight(); y1++)
			{
				long elevation1 = inputFile.GetElevationUnscaled(x, y1);
				if(elevation1 == 0)
				{
					// The pixel is black. Find next non-black pixel.
					for(int y2 = y1 + 1; y2 < inputFile.GetImageHeight(); y2++)
					{
						long elevation2 = inputFile.GetElevationUnscaled(x, y2);
						if(elevation2 != 0)
						{
							// The pixel is non-black. Interpolate.
							// Get last non-black pixel before the gap.
							long last;
							if(y1 > 0)
							{
								last = inputFile.GetElevationUnscaled(x, y1 - 1);
							}
							else
							{
 								last = elevation2;
 							}
							// elevation2 is the first non-zero elevation after the gap.
							// Interpolate between last and elevation2.
							double dY = y2 - y1 + 1; // The length of the gap.
							double elevation3 = last;
							double dElevation = (elevation2 - elevation3) / dY;
							for(int y3 = y1; y3 < y2; y3++)
							{
								elevation3 += dElevation;
								long elevationExisting = outputFile.GetElevationUnscaled(x, y3);
								long mixElevation = long(elevation3 * weight + elevationExisting * weightComplement + 0.5);
								residual += abs(mixElevation - elevationExisting);
								outputFile.SetElevationUnscaled(x, y3, mixElevation);
							}
							y1 = y2;
							break;
						}
					}
				}
			}
		}
		if(residual == oldResidual)
		{
      message.Format(_T("Relaxation achieved after %d iterations."), repeat);
      addMessage(message);
			break;
		}
    if((repeat % 5) == 4)
    {
      message.Format(_T("Iteration %2d: residual %0.5f"),
                     repeat + 1,
                     fabs(double(residual) - double(oldResidual))/double(oldResidual));
      addMessage(message);
    }
		oldResidual = residual;
	}
  inputFile.Close();
  if(outputFile.Write(outputFilename))
  {
    message.Format(_T("Could not write \"%s\"."), outputFilename);
    addMessage(message);
    return 3;
  }
  CString inputFilenameRenamed = inputFilename + ".rm";
  outputFile.Close();
  if(!MoveFile(inputFilename, inputFilenameRenamed))
  {
    message.Format(_T("Could not rename \"%s\" to \"%s\"."), inputFilename, inputFilenameRenamed);
    addMessage(message);
    return 4;
  }
  if(!MoveFile(outputFilename, inputFilename))
  {
    message.Format(_T("Could not rename \"%s\" to \"%s\"."), outputFilename, inputFilename);
    addMessage(message);
    return 5;
  }
  if(removeIntermediateFiles)
  {
    if(!DeleteFile(inputFilenameRenamed))
    {
      message.Format(_T("Could not delete \"%s\"."), inputFilenameRenamed);
      addMessage(message);
      return 5;
    }
  }
  return 0;
}
